@@ -3,7 +3,7 @@
  *
  * Create a repository schema.
  *
- * Copyright (c) 2009-2016, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2009-2017, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 -- Adjust this setting to control where the objects get created.
@@ -36,6 +36,10 @@ CREATE TABLE statsrepo.instance
 	port				integer NOT NULL,
 	pg_version			text,
 	xlog_file_size		bigint,
+	page_size			integer,
+	page_header_size	smallint,
+	htup_header_size	smallint,
+	item_id_size		smallint,
 	PRIMARY KEY (instid),
 	UNIQUE (name, hostname, port)
 );
@@ -333,23 +337,24 @@ CREATE INDEX statsrepo_function_idx ON statsrepo.function(snapid, dbid);
 
 CREATE TABLE statsrepo.autovacuum
 (
-	instid			bigint,
-	start			timestamptz,
-	database		text,
-	schema			text,
-	"table"			text,
-	index_scans		integer,
-	page_removed	integer,
-	page_remain		integer,
-	tup_removed		bigint,
-	tup_remain		bigint,
-	tup_dead		bigint,
-	page_hit		integer,
-	page_miss		integer,
-	page_dirty		integer,
-	read_rate		double precision,
-	write_rate		double precision,
-	duration		real,
+	instid					bigint,
+	start					timestamptz,
+	database				text,
+	schema					text,
+	"table"					text,
+	index_scans				integer,
+	page_removed			bigint,
+	page_remain				bigint,
+	frozen_skipped_pages	bigint,
+	tup_removed				bigint,
+	tup_remain				bigint,
+	tup_dead				bigint,
+	page_hit				integer,
+	page_miss				integer,
+	page_dirty				integer,
+	read_rate				double precision,
+	write_rate				double precision,
+	duration				real,
 	FOREIGN KEY (instid) REFERENCES statsrepo.instance (instid) ON DELETE CASCADE
 );
 CREATE INDEX statsrepo_autovacuum_idx ON statsrepo.autovacuum(instid, start);
@@ -491,12 +496,25 @@ CREATE TABLE statsrepo.lock
 	blockee_pid			integer,
 	blocker_pid			integer,
 	blocker_gid			text,
+	wait_event_type		text,
+	wait_event			text,
 	duration			interval,
 	blockee_query		text,
 	blocker_query		text,
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
 
+CREATE TABLE statsrepo.bgwriter
+(
+	snapid					bigint,
+	buffers_clean			bigint,
+	maxwritten_clean		bigint,
+	buffers_backend			bigint,
+	buffers_backend_fsync	bigint,
+	buffers_alloc			bigint,
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
 CREATE TABLE statsrepo.replication
 (
 	snapid				bigint,
@@ -609,12 +627,7 @@ LANGUAGE sql;
 
 -- get_version() - version of statsrepo schema
 CREATE FUNCTION statsrepo.get_version() RETURNS text AS
-'SELECT CAST(''30200'' AS TEXT)'
-LANGUAGE sql IMMUTABLE;
-
--- page_size() - page size of the monitored instance
-CREATE FUNCTION statsrepo.page_size() RETURNS integer AS
-'SELECT 8192'
+'SELECT CAST(''30300'' AS TEXT)'
 LANGUAGE sql IMMUTABLE;
 
 -- tps() - transaction per seconds
@@ -888,104 +901,6 @@ $$
 $$
 LANGUAGE sql;
 
--- snapshot list
-CREATE FUNCTION statsrepo.get_snapshot_list(
-	IN  instid		bigint,
-	OUT snapid		bigint,
-	OUT check_box	xml,
-	OUT "timestamp"	timestamp(0),
-	OUT size		numeric,
-	OUT diff_size	numeric,
-	OUT commits		numeric(1000,3),
-	OUT	rollbacks	numeric(1000,3),
-	OUT comment		text
-) RETURNS SETOF record AS
-$$
-	SELECT
-		e.snapid,
-		xmlelement(	name input,
-					xmlattributes(	'checkbox' AS  type,
-									to_char(e.snapid, '00000') AS pos,
-									e.snapid AS value)
-					),
-		time::timestamp(0),
-		round(ed.size / 1024 / 1024, 0),
-		round((ed.size - sd.size) / 1024 / 1024, 0),
-		statsrepo.tps(ed.commits - sd.commits, time - time0),
-		statsrepo.tps(ed.rollbacks - sd.rollbacks, time - time0),
-		comment
-	FROM
-		(SELECT	*,
-				lag(snapid) OVER (ORDER BY snapid) AS snapid0,
-				lag(time) OVER (ORDER BY snapid) AS time0
-		FROM statsrepo.snapshot
-		WHERE instid = $1) e
-		LEFT JOIN
-			(SELECT	snapid,
-					sum(size) AS size,
-					sum(xact_commit) AS commits,
-					sum(xact_rollback) AS rollbacks
-			FROM statsrepo.database GROUP BY snapid) AS ed
-			ON ed.snapid = e.snapid
-		LEFT JOIN
-			(SELECT snapid,
-					sum(size) AS size,
-					sum(xact_commit) AS commits,
-					sum(xact_rollback) AS rollbacks
-			FROM statsrepo.database GROUP BY snapid) AS sd
-			ON sd.snapid = e.snapid0;
-$$
-LANGUAGE sql;
-
--- refine snapshot list
-CREATE FUNCTION statsrepo.get_snapshot_list_refine(
-	IN  begin_snapid	bigint,
-	OUT snapid			bigint,
-	OUT check_box		xml,
-	OUT "timestamp"		timestamp(0),
-	OUT size			numeric,
-	OUT diff_size		numeric,
-	OUT commits			numeric(1000,3),
-	OUT	rollbacks		numeric(1000,3),
-	OUT comment			text
-) RETURNS SETOF record AS
-$$
-	SELECT
-		e.snapid,
-		xmlelement(	name input,
-					xmlattributes(	'checkbox' AS  type,
-									to_char(e.snapid, '00000') AS pos,
-									e.snapid AS value)
-					),
-		time::timestamp(0),
-		round(ed.size / 1024 / 1024, 0),
-		round((ed.size - sd.size) / 1024 / 1024, 0),
-		statsrepo.tps(ed.commits - sd.commits, time - time0),
-		statsrepo.tps(ed.rollbacks - sd.rollbacks, time - time0),
-		comment
-	FROM
-		(SELECT	*,
-				lag(snapid) OVER (ORDER BY snapid) AS snapid0,
-				lag(time) OVER (ORDER BY snapid) AS time0
-		FROM statsrepo.snapshot
-		WHERE instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $1)) e
-		LEFT JOIN
-			(SELECT	snapid,
-					sum(size) AS size,
-					sum(xact_commit) AS commits,
-					sum(xact_rollback) AS rollbacks
-			FROM statsrepo.database GROUP BY snapid) AS ed
-			ON ed.snapid = e.snapid
-		LEFT JOIN
-			(SELECT snapid,
-					sum(size) AS size,
-					sum(xact_commit) AS commits,
-					sum(xact_rollback) AS rollbacks
-			FROM statsrepo.database GROUP BY snapid) AS sd
-			ON sd.snapid = e.snapid0;
-$$
-LANGUAGE sql;
-
 -- get min snapshot id from date
 CREATE FUNCTION statsrepo.get_min_snapid(
 	IN  m_host text,
@@ -2128,14 +2043,16 @@ $$
 			t.schema, 
 	 		t.table, 
 			t.n_live_tup,
-			ceil(t.n_live_tup::real / ((statsrepo.page_size() - 24) * statsrepo.pg_fillfactor(t.reloptions, 0) / 100 /
-				(width + 28)))::bigint AS logical_pages,
-			(t.size + CASE t.toastrelid WHEN 0 THEN 0 ELSE tt.size END) / statsrepo.page_size() AS physical_pages
+			ceil(t.n_live_tup::real / ((i.page_size - i.page_header_size) * statsrepo.pg_fillfactor(t.reloptions, 0) / 100 /
+				(width + i.htup_header_size + i.item_id_size)))::bigint AS logical_pages,
+			(t.size + CASE t.toastrelid WHEN 0 THEN 0 ELSE tt.size END) / i.page_size AS physical_pages
 		 FROM
 		 	statsrepo.tables t
+		 	LEFT JOIN statsrepo.snapshot s ON t.snapid = s.snapid
+		 	LEFT JOIN statsrepo.instance i ON s.instid = i.instid
 		 	LEFT JOIN
 		 		(SELECT
-		 			snapid, dbid, tbl, sum(avg_width)::integer + 7 & ~7 AS width
+		 			snapid, dbid, tbl, (sum(avg_width)::integer + 7) & ~7 AS width
 				 FROM
 				 	statsrepo."column" 
 				 WHERE
@@ -2538,6 +2455,43 @@ CREATE FUNCTION statsrepo.get_query_activity_statements(
 	OUT blk_read_time	numeric,
 	OUT blk_write_time	numeric
 ) RETURNS SETOF record AS
+$$
+DECLARE
+	pg_version_num	integer;
+BEGIN
+	SELECT
+		CASE WHEN split_part(pg_version, '.', 3) != '' THEN
+			(split_part(pg_version, '.', 1)::integer * 100 +
+			 split_part(pg_version, '.', 2)::integer) * 100 +
+			 split_part(pg_version, '.', 3)::integer
+		ELSE
+			(split_part(pg_version, '.', 1)::integer * 100 +
+			 substring(split_part(pg_version, '.', 2) from '^\d+')::integer) * 100
+		END
+	INTO pg_version_num FROM statsrepo.instance
+	WHERE instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $1);
+
+	IF pg_version_num >= 90400 THEN
+		RETURN QUERY SELECT * FROM statsrepo.get_query_activity_statemets_body($1, $2);
+	ELSE
+		RETURN QUERY SELECT * FROM statsrepo.get_query_activity_statemets_body_legacy($1, $2);
+	END IF;
+END;
+$$ LANGUAGE plpgsql;
+
+-- internal function for get_query_activity_statements()
+CREATE FUNCTION statsrepo.get_query_activity_statemets_body(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT rolname			text,
+	OUT datname			name,
+	OUT query			text,
+	OUT calls			bigint,
+	OUT total_time		numeric,
+	OUT time_per_call	numeric,
+	OUT blk_read_time	numeric,
+	OUT blk_write_time	numeric
+) RETURNS SETOF record AS
 $$
 	SELECT
 		t1.rolname::text,
@@ -2562,6 +2516,75 @@ $$
 		 		s.dbid,
 		 		s.userid,
 		 		s.queryid,
+		 		max(s.query) AS query,
+				$1 AS first,
+				max(s.snapid) AS last
+			 FROM
+			 	statsrepo.statement s
+				JOIN statsrepo.snapshot ss ON (ss.snapid = s.snapid)
+			 WHERE
+			 	s.snapid >= $1 AND s.snapid <= $2
+				AND ss.instid = (SELECT instid FROM statsrepo.snapshot ss1 WHERE ss1.snapid = $2)
+			 GROUP BY
+				s.dbid,
+				s.userid,
+				s.queryid
+			) AS reg
+		LEFT JOIN statsrepo.statement st1 ON
+			(st1.dbid = reg.dbid AND st1.userid = reg.userid AND
+			 st1.queryid = reg.queryid AND st1.snapid = reg.first)
+		JOIN statsrepo.statement st2 ON
+			(st2.dbid = reg.dbid AND st2.userid = reg.userid AND
+			 st2.queryid = reg.queryid AND st2.snapid = reg.last)
+		JOIN statsrepo.database db ON
+			(db.snapid = reg.first AND db.dbid = reg.dbid)
+		JOIN statsrepo.role rol ON
+			(rol.snapid = reg.first AND rol.userid = reg.userid)
+	) AS t1
+	WHERE
+		t1.calls > 0 and t1.total_time > 0
+	ORDER BY
+		5 DESC,
+		4 DESC;
+$$
+LANGUAGE sql;
+
+-- internal function for get_query_activity_statements()
+CREATE FUNCTION statsrepo.get_query_activity_statemets_body_legacy(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT rolname			text,
+	OUT datname			name,
+	OUT query			text,
+	OUT calls			bigint,
+	OUT total_time		numeric,
+	OUT time_per_call	numeric,
+	OUT blk_read_time	numeric,
+	OUT blk_write_time	numeric
+) RETURNS SETOF record AS
+$$
+		SELECT
+		t1.rolname::text,
+		t1.dbname::name,
+		t1.query,
+		t1.calls,
+		t1.total_time::numeric(1000, 3),
+		(t1.total_time / t1.calls)::numeric(1000, 3),
+		t1.blk_read_time::numeric(1000, 3),
+		t1.blk_write_time::numeric(1000, 3)
+	FROM
+		(SELECT
+			rol.name AS rolname,
+			db.name AS dbname,
+			reg.query,
+			statsrepo.sub(st2.calls, st1.calls) AS calls,
+			statsrepo.sub(st2.total_time, st1.total_time) AS total_time,
+			statsrepo.sub(st2.blk_read_time, st1.blk_read_time) AS blk_read_time,
+			statsrepo.sub(st2.blk_write_time, st1.blk_write_time) AS blk_write_time
+		 FROM
+		 	(SELECT
+		 		s.dbid,
+		 		s.userid,
 		 		s.query,
 				$1 AS first,
 				max(s.snapid) AS last
@@ -2574,16 +2597,13 @@ $$
 			 GROUP BY
 				s.dbid,
 				s.userid,
-				s.queryid,
 				s.query
 			) AS reg
 		LEFT JOIN statsrepo.statement st1 ON
 			(st1.dbid = reg.dbid AND st1.userid = reg.userid AND
-			 st1.queryid = reg.queryid AND
 			 st1.query = reg.query AND st1.snapid = reg.first)
 		JOIN statsrepo.statement st2 ON
 			(st2.dbid = reg.dbid AND st2.userid = reg.userid AND
-			 st2.queryid = reg.queryid AND
 			 st2.query = reg.query AND st2.snapid = reg.last)
 		JOIN statsrepo.database db ON
 			(db.snapid = reg.first AND db.dbid = reg.dbid)
@@ -2759,9 +2779,23 @@ $$
 			(db.snapid = reg.last AND db.dbid = reg.dbid)
 		JOIN statsrepo.role rol ON
 			(rol.snapid = reg.last AND rol.userid = reg.userid)
-		JOIN statsrepo.statement st ON
-			(st.snapid = reg.last AND st.queryid = reg.queryid AND
-			 st.dbid = reg.dbid AND st.userid = reg.userid)
+		JOIN
+			(SELECT queryid,
+					dbid,
+					userid,
+					max(query) AS query
+			 FROM
+			 	statsrepo.statement s
+				JOIN statsrepo.snapshot ss ON (ss.snapid = s.snapid)
+			 WHERE
+			 	s.snapid >= $1 AND s.snapid <= $2
+				AND ss.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+			 GROUP BY
+			 	queryid, dbid, userid
+			) AS st ON
+				st.queryid = reg.queryid AND
+				st.dbid = reg.dbid AND
+				st.userid = reg.userid
 	) AS t1
 	WHERE
 		t1.calls > 0 and t1.total_time > 0
@@ -2839,45 +2873,77 @@ $$
 $$
 LANGUAGE sql;
 
--- generate information that corresponds to 'Replication Activity'
-CREATE FUNCTION statsrepo.get_replication_activity(
-	IN snapid_begin			bigint,
-	IN snapid_end			bigint,
-	OUT usename				name,
-	OUT application_name	text,
-	OUT client_addr			inet,
-	OUT client_hostname		text,
-	OUT client_port			integer,
-	OUT backend_start		timestamp,
-	OUT state				text,
-	OUT current_location	text,
-	OUT sent_location		text,
-	OUT write_location		text,
-	OUT flush_location		text,
-	OUT replay_location		text,
-	OUT sync_priority		integer,
-	OUT sync_state			text
+-- generate information that corresponds to 'BGWriter Statistics'
+CREATE FUNCTION statsrepo.get_bgwriter_tendency(
+	IN snapid_begin				bigint,
+	IN snapid_end				bigint,
+	OUT "timestamp"				text,
+	OUT bgwriter_write_tps		numeric,
+	OUT backend_write_tps		numeric,
+	OUT backend_fsync_tps		numeric,
+	OUT bgwriter_stopscan_tps	numeric,
+	OUT buffer_alloc_tps		numeric
 ) RETURNS SETOF record AS
 $$
 	SELECT
-		usename,
-		application_name,
-		client_addr,
-		client_hostname,
-		client_port,
-		backend_start::timestamp(0),
-		state,
-		current_location,
-		sent_location,
-		write_location,
-		flush_location,
-		replay_location,
-		sync_priority,
-		sync_state
+		t.timestamp,
+		statsrepo.tps(t.bgwriter_write, t.duration),
+		statsrepo.tps(t.backend_write, t.duration),
+		statsrepo.tps(t.backend_fsync, t.duration),
+		statsrepo.tps(t.bgwriter_stopscan, t.duration),
+		statsrepo.tps(t.buffer_alloc, t.duration)
 	FROM
-		statsrepo.replication
+	(
+		SELECT
+			s.snapid,
+			to_char(s.time, 'YYYY-MM-DD HH24:MI') AS timestamp,
+			b.buffers_clean - lag(b.buffers_clean) OVER w AS bgwriter_write,
+			b.buffers_backend - lag(b.buffers_backend) OVER w AS backend_write,
+			b.buffers_backend_fsync - lag(b.buffers_backend_fsync) OVER w AS backend_fsync,
+			b.maxwritten_clean - lag(b.maxwritten_clean) OVER w AS bgwriter_stopscan,
+			b.buffers_alloc - lag(b.buffers_alloc) OVER w AS buffer_alloc,
+			s.time - lag(s.time) OVER w AS duration
+		FROM
+			statsrepo.bgwriter b,
+			statsrepo.snapshot s
+		WHERE
+			b.snapid BETWEEN $1 AND $2
+			AND b.snapid = s.snapid
+			AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+		WINDOW w AS (ORDER BY s.snapid)
+		ORDER BY
+			s.snapid
+	) t
 	WHERE
-		snapid = $2
+		t.snapid > $1;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'BGWriter Statistics'
+CREATE OR REPLACE FUNCTION statsrepo.get_bgwriter_stats(
+	IN snapid_begin				bigint,
+	IN snapid_end				bigint,
+	OUT bgwriter_write_avg		numeric,
+	OUT bgwriter_write_max		numeric,
+	OUT backend_write_avg		numeric,
+	OUT backend_write_max		numeric,
+	OUT backend_fsync_avg		numeric,
+	OUT backend_fsync_max		numeric,
+	OUT bgwriter_stopscan_avg	numeric,
+	OUT buffer_alloc_avg		numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		round(avg(bgwriter_write_tps), 3),
+		round(max(bgwriter_write_tps), 3),
+		round(avg(backend_write_tps), 3),
+		round(max(backend_write_tps), 3),
+		round(avg(backend_fsync_tps), 3),
+		round(max(backend_fsync_tps), 3),
+		round(avg(bgwriter_stopscan_tps), 3),
+		round(avg(buffer_alloc_tps), 3)
+	FROM
+		statsrepo.get_bgwriter_tendency($1, $2);
 $$
 LANGUAGE sql;
 
@@ -2903,8 +2969,7 @@ $$
 			split_part(r.current_location, ' ', 1),
 			split_part(r.replay_location, ' ', 1),
 			i.xlog_file_size),
-		(SELECT sync_state FROM statsrepo.replication WHERE snapid = $2
-			AND client_addr = r.client_addr AND client_port = r.client_port)
+		r.sync_state
 	FROM
 		statsrepo.replication r,
 		statsrepo.snapshot s,
@@ -2921,6 +2986,61 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Replication Activity'
+CREATE FUNCTION statsrepo.get_replication_activity(
+	IN snapid_begin			bigint,
+	IN snapid_end			bigint,
+	OUT usename				name,
+	OUT application_name	text,
+	OUT client_addr			inet,
+	OUT client_hostname		text,
+	OUT client_port			integer,
+	OUT backend_start		timestamp,
+	OUT state				text,
+	OUT current_location	text,
+	OUT sent_location		text,
+	OUT write_location		text,
+	OUT flush_location		text,
+	OUT replay_location		text,
+	OUT sync_priority		integer,
+	OUT sync_state			text,
+	OUT replay_delay_avg	numeric,
+	OUT replay_delay_peak	numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		r.usename,
+		r.application_name,
+		r.client_addr,
+		r.client_hostname,
+		r.client_port,
+		r.backend_start::timestamp(0),
+		r.state,
+		r.current_location,
+		r.sent_location,
+		r.write_location,
+		r.flush_location,
+		r.replay_location,
+		r.sync_priority,
+		r.sync_state,
+		d.replay_delay_avg,
+		d.replay_delay_peak
+	FROM
+		statsrepo.replication r,
+		(SELECT
+			client,
+			avg(replay_delay_size) AS replay_delay_avg,
+			max(replay_delay_size) AS replay_delay_peak
+		 FROM
+		 	statsrepo.get_replication_delays($1, $2)
+		 GROUP BY
+			client) d
+	WHERE
+		d.client = (host(r.client_addr) || ':' || r.client_port)
+		AND r.snapid = $2;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Setting Parameters'
 CREATE FUNCTION statsrepo.get_setting_parameters(
 	IN snapid_begin	bigint,
